﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Text;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.ConditionActionEncoding;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Encodings.VariableVector {
  [StorableClass]
  [Item("VariableVectorCondition", "")]
  public class VariableVectorCondition : Dictionary<string, IVariable>, ICondition {

    public IOrderedEnumerable<string> Order { get { return this.Keys.OrderBy(x => x); } }

    public int VirtualLength { get { return this.Values.Sum(x => x.VirtualLength); } }

    [StorableConstructor]
    protected VariableVectorCondition(bool deserializing) { }
    protected VariableVectorCondition(VariableVectorCondition original, Cloner cloner) {
      foreach (var item in original) {
        Add((string)item.Key.Clone(), (IVariable)item.Value.Clone());
      }
    }
    public VariableVectorCondition() : base() { }
    public VariableVectorCondition(int capacity) : base(capacity) { }
    public VariableVectorCondition(IDictionary<string, IVariable> dictionary) : base(dictionary) { }

    public bool Match(IInput target) {
      var targetCast = target as VariableVectorInput;
      if (targetCast == null) { return false; }
      if (!this.Keys.All(x => targetCast.Keys.Contains(x))) { return false; }
      foreach (var key in this.Keys) {
        if (!this[key].MatchInput(targetCast[key])) {
          return false;
        }
      }
      return true;
    }

    public void Add(IEnumerable<IVariable> condition) {
      foreach (var variable in condition) {
        this.Add(variable.VariableName, variable);
      }
    }

    public override string ToString() {
      StringBuilder sb = new StringBuilder();
      bool first = true;
      foreach (var variable in this.Values) {
        if (first) {
          first = false;
        } else {
          sb.Append(";");
        }
        sb.Append(variable.ToString());
      }
      return sb.ToString();
    }

    public void Randomize(IRandom random, double spreadPercentage) {
      foreach (var variable in this.Values) {
        if (variable is DoubleVariable) {
          ((DoubleVariable)variable).Randomize(random, spreadPercentage);
        } else {
          variable.Randomize(random);
        }
      }
    }

    public bool Identical(VariableVectorCondition target) {
      if (!this.Order.SequenceEqual(target.Order)) { return false; }
      foreach (var keyValuePair in target) {
        if (!this[keyValuePair.Key].Identical(keyValuePair.Value)) {
          return false;
        }
      }
      return true;
    }

    public virtual string ItemName {
      get { return ItemAttribute.GetName(this.GetType()); }
    }
    public virtual string ItemDescription {
      get { return ItemAttribute.GetDescription(this.GetType()); }
    }
    public Version ItemVersion {
      get { return ItemAttribute.GetVersion(this.GetType()); }
    }
    public static Image StaticItemImage {
      get { return HeuristicLab.Common.Resources.VSImageLibrary.Class; }
    }
    public virtual Image ItemImage {
      get { return ItemAttribute.GetImage(this.GetType()); }
    }

    public virtual IDeepCloneable Clone(Cloner cloner) {
      return new VariableVectorCondition(this, cloner);
    }

    public object Clone() {
      return Clone(new Cloner());
    }

    public event EventHandler ItemImageChanged;
    protected virtual void OnItemImageChanged() {
      EventHandler handler = ItemImageChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    public event EventHandler ToStringChanged;
    protected virtual void OnToStringChanged() {
      EventHandler handler = ToStringChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }
  }
}
